@@ -66,6 +66,7 @@
 #define CONF_KEY "show_notification_icon"
 
 #define VIDEO_KEYSYM    "XF86Display"
+#define ROTATE_KEYSYM   "XF86RotateWindows"
 
 /* Number of seconds that the confirmation dialog will last before it resets the
  * RANDR configuration to its old state.
@@ -87,8 +88,12 @@ struct GsdXrandrManagerPrivate
 {
         DBusGConnection *dbus_connection;
 
-        /* Key code of the fn-F7 video key (XF86Display) */
-        guint keycode;
+        /* Key code of the XF86Display key (Fn-F7 on Thinkpads, Fn-F4 on HP machines, etc.) */
+        guint switch_video_mode_keycode;
+
+        /* Key code of the XF86RotateWindows key (present on some tablets) */
+        guint rotate_windows_keycode;
+
         GnomeRRScreen *rw_screen;
         gboolean running;
 
@@ -107,6 +112,14 @@ struct GsdXrandrManagerPrivate
         guint32 last_config_timestamp;
 };
 
+static const GnomeRRRotation possible_rotations[] = {
+        GNOME_RR_ROTATION_0,
+        GNOME_RR_ROTATION_90,
+        GNOME_RR_ROTATION_180,
+        GNOME_RR_ROTATION_270
+        /* We don't allow REFLECT_X or REFLECT_Y for now, as gnome-display-properties doesn't allow them, either */
+};
+
 static void     gsd_xrandr_manager_class_init  (GsdXrandrManagerClass *klass);
 static void     gsd_xrandr_manager_init        (GsdXrandrManager      *xrandr_manager);
 static void     gsd_xrandr_manager_finalize    (GObject             *object);
@@ -115,6 +128,11 @@ static void error_message (GsdXrandrManager *mgr, const char *primary_text, GErr
 
 static void status_icon_popup_menu (GsdXrandrManager *manager, guint button, guint32 timestamp);
 static void run_display_capplet (GtkWidget *widget);
+static void get_allowed_rotations_for_output (GnomeRRConfig *config,
+                                              GnomeRRScreen *rr_screen,
+                                              GnomeOutputInfo *output,
+                                              int *out_num_rotations,
+                                              GnomeRRRotation *out_rotations);
 
 G_DEFINE_TYPE (GsdXrandrManager, gsd_xrandr_manager, G_TYPE_OBJECT)
 
@@ -148,7 +166,9 @@ show_timestamps_dialog (GsdXrandrManager *manager, const char *msg)
 #endif
 }
 
-/* Optionally filters out GNOME_RR_ERROR_NO_MATCHING_CONFIG from
+/* This function centralizes the use of gnome_rr_config_apply_from_filename_with_time().
+ *
+ * Optionally filters out GNOME_RR_ERROR_NO_MATCHING_CONFIG from
  * gnome_rr_config_apply_from_filename_with_time(), since that is not usually an error.
  */
 static gboolean
@@ -188,6 +208,28 @@ fail:
         return FALSE;
 }
 
+/* This function centralizes the use of gnome_rr_config_apply_with_time().
+ *
+ * Applies a configuration and displays an error message if an error happens.
+ * We just return whether setting the configuration succeeded.
+ */
+static gboolean
+apply_configuration_and_display_error (GsdXrandrManager *manager, GnomeRRConfig *config, guint32 timestamp)
+{
+        GsdXrandrManagerPrivate *priv = manager->priv;
+        GError *error;
+        gboolean success;
+
+        error = NULL;
+        success = gnome_rr_config_apply_with_time (config, priv->rw_screen, timestamp, &error);
+        if (!success) {
+                error_message (manager, _("Could not switch the monitor configuration"), error, NULL);
+                g_error_free (error);
+        }
+
+        return success;
+}
+
 static void
 restore_backup_configuration_without_messages (const char *backup_filename, const char *intended_filename)
 {
@@ -450,8 +492,7 @@ is_laptop (GnomeOutputInfo *output)
         if (output->connected && output_name &&
             (strstr (output_name, "lvds")	||
              strstr (output_name, "LVDS")	||
-             strstr (output_name, "Lvds")       ||
-             strstr (output_name, "LCD")))
+             strstr (output_name, "Lvds")))
         {
                 return TRUE;
         }
@@ -656,10 +697,12 @@ make_laptop_setup (GnomeRRScreen *screen)
                 GnomeOutputInfo *info = result->outputs[i];
 
                 if (is_laptop (info)) {
-                        if (!turn_on (screen, info, 0, 0)) {
-                                gnome_rr_config_free (result);
-                                result = NULL;
-                                break;
+                        if (!info->on) {
+                                if (!turn_on (screen, info, 0, 0)) {
+                                        gnome_rr_config_free (result);
+                                        result = NULL;
+                                        break;
+                                }
                         }
                 }
                 else {
@@ -679,8 +722,14 @@ make_laptop_setup (GnomeRRScreen *screen)
 static int
 turn_on_and_get_rightmost_offset (GnomeRRScreen *screen, GnomeOutputInfo *info, int x)
 {
-        if (turn_on (screen, info, x, 0))
+        if (info->on) {
+                info->x = x;
+                info->y = 0;
                 x += info->width;
+        } else {
+                if (turn_on (screen, info, x, 0))
+                        x += info->width;
+        }
 
         return x;
 }
@@ -732,8 +781,9 @@ make_other_setup (GnomeRRScreen *screen)
                         info->on = FALSE;
                 }
                 else {
-                        if (info->connected)
+                        if (info->connected && !info->on) {
                                 turn_on (screen, info, 0, 0);
+                        }
                }
         }
 
@@ -969,11 +1019,7 @@ handle_fn_f7 (GsdXrandrManager *mgr, guint32 timestamp)
 
                 g_debug ("applying");
 
-                error = NULL;
-                if (!gnome_rr_config_apply_with_time (priv->fn_f7_configs[mgr->priv->current_fn_f7_config], screen, timestamp, &error)) {
-                        error_message (mgr, _("Could not switch the monitor configuration"), error, NULL);
-                        g_error_free (error);
-                }
+                apply_configuration_and_display_error (mgr, priv->fn_f7_configs[mgr->priv->current_fn_f7_config], timestamp);
         }
         else {
                 g_debug ("no configurations generated");
@@ -981,6 +1027,113 @@ handle_fn_f7 (GsdXrandrManager *mgr, guint32 timestamp)
         g_debug ("done handling fn-f7");
 }
 
+static GnomeOutputInfo *
+get_laptop_output_info (GnomeRRConfig *config)
+{
+        int i;
+
+        for (i = 0; config->outputs[i] != NULL; i++) {
+                GnomeOutputInfo *info;
+
+                info = config->outputs[i];
+                if (is_laptop (info))
+                        return info;
+        }
+
+        return NULL;
+
+}
+
+static GnomeRRRotation
+get_next_rotation (GnomeRRRotation allowed_rotations, GnomeRRRotation current_rotation)
+{
+        int i;
+        int current_index;
+
+        /* First, find the index of the current rotation */
+
+        current_index = -1;
+
+        for (i = 0; i < G_N_ELEMENTS (possible_rotations); i++) {
+                GnomeRRRotation r;
+
+                r = possible_rotations[i];
+                if (r == current_rotation) {
+                        current_index = i;
+                        break;
+                }
+        }
+
+        if (current_index == -1) {
+                /* Huh, the current_rotation was not one of the supported rotations.  Bail out. */
+                return current_rotation;
+        }
+
+        /* Then, find the next rotation that is allowed */
+
+        i = (current_index + 1) % G_N_ELEMENTS (possible_rotations);
+
+        while (1) {
+                GnomeRRRotation r;
+
+                r = possible_rotations[i];
+                if (r == current_rotation) {
+                        /* We wrapped around and no other rotation is suported.  Bummer. */
+                        return current_rotation;
+                } else if (r & allowed_rotations)
+                        return r;
+
+                i = (i + 1) % G_N_ELEMENTS (possible_rotations);
+        }
+}
+
+/* We use this when the XF86RotateWindows key is pressed.  That key is present
+ * on some tablet PCs; they use it so that the user can rotate the tablet
+ * easily.
+ */
+static void
+handle_rotate_windows (GsdXrandrManager *mgr, guint32 timestamp)
+{
+        GsdXrandrManagerPrivate *priv = mgr->priv;
+        GnomeRRScreen *screen = priv->rw_screen;
+        GnomeRRConfig *current;
+        GnomeOutputInfo *rotatable_output_info;
+        int num_allowed_rotations;
+        GnomeRRRotation allowed_rotations;
+        GnomeRRRotation next_rotation;
+
+        g_debug ("Handling XF86RotateWindows");
+
+        /* Which output? */
+
+        current = gnome_rr_config_new_current (screen);
+
+        rotatable_output_info = get_laptop_output_info (current);
+        if (rotatable_output_info == NULL) {
+                g_debug ("No laptop outputs found to rotate; XF86RotateWindows key will do nothing");
+                goto out;
+        }
+
+        /* Which rotation? */
+
+        get_allowed_rotations_for_output (current, priv->rw_screen, rotatable_output_info, &num_allowed_rotations, &allowed_rotations);
+        next_rotation = get_next_rotation (allowed_rotations, rotatable_output_info->rotation);
+
+        if (next_rotation == rotatable_output_info->rotation) {
+                g_debug ("No rotations are supported other than the current one; XF86RotateWindows key will do nothing");
+                goto out;
+        }
+
+        /* Rotate */
+
+        rotatable_output_info->rotation = next_rotation;
+
+        apply_configuration_and_display_error (mgr, current, timestamp);
+
+out:
+        gnome_rr_config_free (current);
+}
+
 static GdkFilterReturn
 event_filter (GdkXEvent           *xevent,
               GdkEvent            *event,
@@ -996,8 +1149,11 @@ event_filter (GdkXEvent           *xevent,
         if (xev->xany.type != KeyPress && xev->xany.type != KeyRelease)
                 return GDK_FILTER_CONTINUE;
 
-        if (xev->xany.type == KeyPress && xev->xkey.keycode == manager->priv->keycode) {
-                handle_fn_f7 (manager, xev->xkey.time);
+        if (xev->xany.type == KeyPress) {
+                if (xev->xkey.keycode == manager->priv->switch_video_mode_keycode)
+                        handle_fn_f7 (manager, xev->xkey.time);
+                else if (xev->xkey.keycode == manager->priv->rotate_windows_keycode)
+                        handle_rotate_windows (manager, xev->xkey.time);
 
                 return GDK_FILTER_CONTINUE;
         }
@@ -1134,13 +1290,8 @@ auto_configure_outputs (GsdXrandrManager *manager, guint32 timestamp)
 
         /* Apply the configuration! */
 
-        if (applicable) {
-                error = NULL;
-                if (!gnome_rr_config_apply_with_time (config, priv->rw_screen, timestamp, &error)) {
-                        error_message (manager, _("Could not switch the monitor configuration"), error, NULL);
-                        g_error_free (error);
-                }
-        }
+        if (applicable)
+                apply_configuration_and_display_error (manager, config, timestamp);
 
         g_list_free (just_turned_on);
         gnome_rr_config_free (config);
@@ -1169,6 +1320,23 @@ auto_configure_outputs (GsdXrandrManager *manager, guint32 timestamp)
 #endif
 }
 
+static void
+apply_color_profiles (void)
+{
+        gboolean ret;
+        GError *error = NULL;
+
+        /* run the gnome-color-manager apply program */
+        ret = g_spawn_command_line_async (BINDIR "/gcm-apply", &error);
+        if (!ret) {
+                /* only print the warning if the binary is installed */
+                if (error->code != G_SPAWN_ERROR_NOENT) {
+                        g_warning ("failed to apply color profiles: %s", error->message);
+                }
+                g_error_free (error);
+        }
+}
+
 static void
 on_randr_event (GnomeRRScreen *screen, gpointer data)
 {
@@ -1260,6 +1428,9 @@ on_randr_event (GnomeRRScreen *screen, gpointer data)
 #endif
         }
 
+        /* poke gnome-color-manager */
+        apply_color_profiles ();
+
         refresh_tray_icon_menu_if_active (manager, MAX (change_timestamp, config_timestamp));
 }
 
@@ -1477,17 +1648,12 @@ make_menu_item_for_output_title (GsdXrandrManager *manager, GnomeOutputInfo *out
 }
 
 static void
-get_allowed_rotations_for_output (GsdXrandrManager *manager, GnomeOutputInfo *output, int *out_num_rotations, GnomeRRRotation *out_rotations)
+get_allowed_rotations_for_output (GnomeRRConfig *config,
+                                  GnomeRRScreen *rr_screen,
+                                  GnomeOutputInfo *output,
+                                  int *out_num_rotations,
+                                  GnomeRRRotation *out_rotations)
 {
-        static const GnomeRRRotation possible_rotations[] = {
-                GNOME_RR_ROTATION_0,
-                GNOME_RR_ROTATION_90,
-                GNOME_RR_ROTATION_180,
-                GNOME_RR_ROTATION_270
-                /* We don't allow REFLECT_X or REFLECT_Y for now, as gnome-display-properties doesn't allow them, either */
-        };
-
-        struct GsdXrandrManagerPrivate *priv = manager->priv;
         GnomeRRRotation current_rotation;
         int i;
 
@@ -1505,7 +1671,7 @@ get_allowed_rotations_for_output (GsdXrandrManager *manager, GnomeOutputInfo *ou
 
                 output->rotation = rotation_to_test;
 
-                if (gnome_rr_config_applicable (priv->configuration, priv->rw_screen, NULL)) { /* NULL-GError */
+                if (gnome_rr_config_applicable (config, rr_screen, NULL)) { /* NULL-GError */
                         (*out_num_rotations)++;
                         (*out_rotations) |= rotation_to_test;
                 }
@@ -1668,10 +1834,11 @@ add_items_for_rotations (GsdXrandrManager *manager, GnomeOutputInfo *output, Gno
 static void
 add_rotation_items_for_output (GsdXrandrManager *manager, GnomeOutputInfo *output)
 {
+        struct GsdXrandrManagerPrivate *priv = manager->priv;
         int num_rotations;
         GnomeRRRotation rotations;
 
-        get_allowed_rotations_for_output (manager, output, &num_rotations, &rotations);
+        get_allowed_rotations_for_output (priv->configuration, priv->rw_screen, output, &num_rotations, &rotations);
 
         if (num_rotations == 1)
                 add_unsupported_rotation_item (manager);
@@ -1912,11 +2079,23 @@ gsd_xrandr_manager_start (GsdXrandrManager *manager,
                         (GConfClientNotifyFunc)on_config_changed,
                         manager, NULL, NULL);
 
-        if (manager->priv->keycode) {
+        if (manager->priv->switch_video_mode_keycode) {
                 gdk_error_trap_push ();
 
                 XGrabKey (gdk_x11_get_default_xdisplay(),
-                          manager->priv->keycode, AnyModifier,
+                          manager->priv->switch_video_mode_keycode, AnyModifier,
+                          gdk_x11_get_default_root_xwindow(),
+                          True, GrabModeAsync, GrabModeAsync);
+
+                gdk_flush ();
+                gdk_error_trap_pop ();
+        }
+
+        if (manager->priv->rotate_windows_keycode) {
+                gdk_error_trap_push ();
+
+                XGrabKey (gdk_x11_get_default_xdisplay(),
+                          manager->priv->rotate_windows_keycode, AnyModifier,
                           gdk_x11_get_default_root_xwindow(),
                           True, GrabModeAsync, GrabModeAsync);
 
@@ -1945,13 +2124,25 @@ gsd_xrandr_manager_stop (GsdXrandrManager *manager)
 
         manager->priv->running = FALSE;
 
-        gdk_error_trap_push ();
+        if (manager->priv->switch_video_mode_keycode) {
+                gdk_error_trap_push ();
 
-        XUngrabKey (gdk_x11_get_default_xdisplay(),
-                    manager->priv->keycode, AnyModifier,
-                    gdk_x11_get_default_root_xwindow());
+                XUngrabKey (gdk_x11_get_default_xdisplay(),
+                            manager->priv->switch_video_mode_keycode, AnyModifier,
+                            gdk_x11_get_default_root_xwindow());
 
-        gdk_error_trap_pop ();
+                gdk_error_trap_pop ();
+        }
+
+        if (manager->priv->rotate_windows_keycode) {
+                gdk_error_trap_push ();
+
+                XUngrabKey (gdk_x11_get_default_xdisplay(),
+                            manager->priv->rotate_windows_keycode, AnyModifier,
+                            gdk_x11_get_default_root_xwindow());
+
+                gdk_error_trap_pop ();
+        }
 
         gdk_window_remove_filter (gdk_get_default_root_window (),
                                   (GdkFilterFunc) event_filter,
@@ -2060,16 +2251,25 @@ gsd_xrandr_manager_class_init (GsdXrandrManagerClass *klass)
         g_type_class_add_private (klass, sizeof (GsdXrandrManagerPrivate));
 }
 
+static guint
+get_keycode_for_keysym_name (const char *name)
+{
+        Display *dpy;
+        guint keyval;
+
+        dpy = gdk_x11_get_default_xdisplay ();
+
+        keyval = gdk_keyval_from_name (name);
+        return XKeysymToKeycode (dpy, keyval);
+}
+
 static void
 gsd_xrandr_manager_init (GsdXrandrManager *manager)
 {
-        Display *dpy = gdk_x11_get_default_xdisplay ();
-        guint keyval = gdk_keyval_from_name (VIDEO_KEYSYM);
-        guint keycode = XKeysymToKeycode (dpy, keyval);
-
         manager->priv = GSD_XRANDR_MANAGER_GET_PRIVATE (manager);
 
-        manager->priv->keycode = keycode;
+        manager->priv->switch_video_mode_keycode = get_keycode_for_keysym_name (VIDEO_KEYSYM);
+        manager->priv->rotate_windows_keycode = get_keycode_for_keysym_name (ROTATE_KEYSYM);
 
         manager->priv->current_fn_f7_config = -1;
         manager->priv->fn_f7_configs = NULL;
